# -*- coding: utf-8 -*-
"""
ImageDisplay is a widget for PyQT which provides a scientific image display
panel.

ImageDisplay is based on a QLabel widget.

@author: Mike Hughes
Applied Optics Group
School of Physics & Astronomoy
University of Kent
"""

from PyQt5 import QtGui, QtCore, QtWidgets
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from PyQt5.QtWidgets import *
from PyQt5.QtGui import QPalette, QColor, QImage, QPixmap, QPainter, QPen, QGuiApplication
from PyQt5.QtGui import QPainter, QBrush, QPen, QFont, QFontMetrics

import numpy as np
from matplotlib import cm
import math
import time

import cv2 as cv

class ImageDisplay(QLabel):
    
   ELLIPSE = 0
   LINE = 1
   POINT = 2
   RECTANGLE = 3
   TEXT = 4
  
   MONO = 0
   RGB = 1
   GRAPH = 2
    
   mouseMoved = pyqtSignal(int, int)
   roiChanged = pyqtSignal()  


   imageSize = (0,0)
   
   isStatusBar = True
   autoScale = True
   enableZoom = True
   isRoiEnabled = True
   isZoomEnabled = True
   isZoomIndicator = True
   
   panning = False
   zoomLocation = None
   pmap = None
   dragging = False
   dragToX = None
   dragToY = None
   imageWidth = None
   imageHeight = None
   
   
   graph_line_width = 5
   graph_width = 512
   graph_height = 512
   displayMin = 0
   displayMax = 255
   zoomLevel = 0
   zoomStepDivider = 2   
    
   zoomIndicatorPen = QPen(Qt.white, 1, Qt.SolidLine)
   zoomIndicatorBrush = QBrush(Qt.white, Qt.SolidPattern)
   zoomIndicatorWidth = 60
   zoomIndicatorOffsetX = 20
   zoomIndicatorOffsetY = 20
   
   roiDragContrastPen = QPen(Qt.white, 2, Qt.SolidLine)
   roiDragPen = QPen(Qt.red, 2, Qt.DotLine)
   roiContrastPen = QPen(Qt.white, 2, Qt.SolidLine)
   roiPen = QPen(Qt.green, 2, Qt.DotLine)
   
   statusPen =  QPen(Qt.white, 2, Qt.SolidLine)
   statusBrush = QBrush(Qt.white, Qt.SolidPattern)
   statusTextPen = QPen(Qt.black, 2, Qt.SolidLine)
   

   graphPen = QPen(Qt.white, 1, Qt.SolidLine)
   graphCursorBrush = QBrush(Qt.white, Qt.SolidPattern)
   graphCursorPen = QPen(Qt.white, 2, Qt.SolidLine)
   graphCursorSize = 4
   graphLabelPen = QPen(Qt.white, 2, Qt.SolidLine)
   graphZeroPen = QPen(Qt.white, 1, Qt.DotLine)
   graphDisplayMin = None
   graphDisplayMax = None
   
   imageMode = MONO
   
   minAutoscaleUpper = -1e10
   
   overlays = []
   nOverlays = 0
   
   colortable = None
   roi = None
   
   lastFrameTime = 0


   
   
   def __init__(self, **kwargs):
       
       self.name = kwargs.get('name', 'noname')
       
       super().__init__()

       self.setMouseTracking(True)
       self.setCursor(Qt.CrossCursor)    
       self.mouseX = 0
       self.mouseY = 0
       
       self.set_image(np.zeros((20,20)))      
       #self.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding,  QtWidgets.QSizePolicy.MinimumExpanding))
       self.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Ignored,  QtWidgets.QSizePolicy.Ignored))

       self.setMaximumSize(2048, 2048)
       self.setMinimumSize(400, 400)

       
       self.setAlignment(Qt.AlignTop)
       self.installEventFilter(self)
       self.setStyleSheet("border:1px solid white")
       self.setAlignment(Qt.AlignCenter)
       
       #self.setSizePolicy(QSizePolicy.Ignored)
       
   
   def set_name(name):
       self.name = name
       
   
   def mouseMoveEvent(self, event):
       """ Record mouse positions and update status bar  """  

       if self.pmap is not None:
           
           self.mouseX, self.mouseY = self.image_coords(event.x(), event.y()) 


           if self.panning and self.is_mouse_on_image():
            
              newX, newY = self.mouseX, self.mouseY
              moveX = newX - self.panningX
              moveY = newY - self.panningY
              h,w = self.currentImage.shape[0:2]
            
              self.displayX = min(max(self.displayX - moveX,0), w - self.displayW)
              self.displayY = min(max(self.displayY - moveY,0), h - self.displayH)
              self.panningX, self.panningY = self.image_coords(event.x(), event.y())

              self.set_image(self.currentImage)
         
           if self.is_mouse_on_image():
               closestX,closestY = self.image_coords_closest(event.x(), event.y()) #self.mouseMoved.emit(self.mouseX, self.mouseY)
               self.dragToX = closestX
               self.dragToY = closestY
           else:
               self.mouseX = None
               self.mouseY = None
       self.update()     
           
          
   def mousePressEvent(self, event):
       """Start of ROI dragging """
       if event.buttons() == QtCore.Qt.LeftButton and self.isRoiEnabled and self.mouseX is not None and self.mouseY is not None:
           self.dragging = True
           self.roi = None
           self.dragX, self.dragY = self.image_coords_closest(event.x(), event.y()) #self.mouseMoved.emit(self.mouseX, self.mouseY)
           #self.dragX = closestX
           #self.dragY = closestY
           self.dragToX = self.dragX
           self.dragToY = self.dragY
       elif (event.buttons() == QtCore.Qt.MidButton or event.buttons() == QtCore.Qt.RightButton) and self.is_mouse_on_image():
           self.panning = True
           self.panningX, self.panningY = (self.mouseX, self.mouseY)
       
       
   def mouseReleaseEvent(self, event):
       """ End of ROI dragging """  
       if self.dragging:
           self.dragging = False
           if self.dragX != self.dragToX and self.dragY != self.dragToY:
               self.roi = int(round(min(self.dragX, self.dragToX))), int(round(min(self.dragY, self.dragToY))), int(round(max(self.dragX, self.dragToX))), int(round(max(self.dragY, self.dragToY)))
           else:
               self.roi = None
           self.roiChanged.emit()
           
            
       self.panning = False
       
       
   def wheelEvent(self,event):
       """ Updates zoom level when mouse wheel is scrolled"""
       if self.imageMode == self.GRAPH:
           self.zoomLevel = 0
           return

       if self.isZoomEnabled and self.mouseX:
           self.zoomLocation = (self.mouseX, self.mouseY)           
           self.zoomLevel = self.zoomLevel + round(event.angleDelta().y() / 120) / self.zoomStepDivider
           self.zoomLevel = max(self.zoomLevel, 0)
           self.set_image(self.currentImage)
  
           
   def is_mouse_on_image(self):
       """ Returns True if mouse if over image"""
       if self.mouseX is None or self.mouseY is None:
           
           return False
       
       if self.mouseX >=0 and self.mouseY >=0 and self.mouseX < self.imageSize[1] and self.mouseY < self.imageSize[0]:
           return True
       else:
           return False
       
        
   def set_graph(self, graph):
       self.graphData = graph       
       
       img = np.zeros((self.graph_width, self.graph_height))
       self.set_image(img)
       
       self.imageMode = self.GRAPH
       self.zoomLevel = 0



       
       
       
   def set_image(self,img):
       """ Sets the image `img` as the current image. `img` is a numpy array, if
       it has three dimensions then it is assumed that it is a colour images, and
       if it has two dimensions then it is assumed that is is a monochrome image."""
      
           
       if img.ndim > 2:
           self.set_rgb_image(img)
       else:
           self.set_mono_image(img)
           
           
   def set_mono_image(self, img):
       """ Sets a grayscale image as the current image """
       
       # Compare width and height with width and height of last image. If it
       # has changed then we adjust the displayX and displayY so we are looking
       # at the same area.
       if self.imageWidth is not None:       
           if np.shape(img)[1] != self.imageWidth:
               self.displayX = int(self.displayX / self.imageWidth * np.shape(img)[1])
           if np.shape(img)[0] != self.imageHeight:
               self.displayY = int(self.displayY / self.imageHeight * np.shape(img)[0])
           
       self.imageWidth = np.shape(img)[1]    
       self.imageHeight = np.shape(img)[0]    
       
       
       
       self.currentImage = img
       self.imageMode = self.MONO
       self.imageSize = np.shape(img)
       if img is not None and np.size(img) > 0:           
           
           t1 = time.perf_counter()
           if self.autoScale: #and np.max(img) != 0:
               img = img.astype('float32')
               img = img - np.min(img)
               m = max(np.max(img), self.minAutoscaleUpper)
               if m > 0:
                   sf =  255 / m
               else:
                   sf = 0
               img = img * sf
              
               #img = cv.normalize(img,0,255,cv.NORM_MINMAX)           
            
           img = img.astype('uint8')
           
           
           #if self.fixedView:
           #    self.getView(img)
               
           #else:    
           self.displayImage = self.zoom(img)          
            
           self.image = QtGui.QImage(self.displayImage.copy(), self.displayImage.shape[1], self.displayImage.shape[0], self.displayImage.shape[1], QtGui.QImage.Format_Indexed8)
                      
           # Set colormap
           if self.colortable is not None:
               self.image.setColorTable(self.colortable)
           
           scaledSize = QtCore.QSize(self.geometry().width(), self.geometry().height()-40)
           self.pmap = QtGui.QPixmap.fromImage(self.image).scaled(scaledSize, QtCore.Qt.KeepAspectRatio)
           self.setPixmap(self.pmap)
      
       else:

           self.pmap = None


   def set_rgb_image(self, img):
       """ Sets a colour RGB image as the current image """
       
       
       # Compare width and height with width and height of last image. If it
       # has changed then we adjust the displayX and displayY so we are looking
       # at the same area.
       if self.imageWidth is not None:       
           if np.shape(img)[1] != self.imageWidth:
               self.displayX = int(self.displayX / self.imageWidth * np.shape(img)[1])
           if np.shape(img)[0] != self.imageHeight:
               self.displayY = int(self.displayY / self.imageHeight * np.shape(img)[0])
           
       self.imageWidth = np.shape(img)[1]    
       self.imageHeight = np.shape(img)[0]    
       
       
       self.currentImage = img.astype(float)       
       self.imageMode = self.RGB       
       self.imageSize = np.shape(img)

       
       if img is not None and np.size(img) > 0:
           
           # img = img.astype('float')
           
           # if self.autoScale and np.max(img) != 0:
           #     img = img - np.min(img)
           #     img = (img / np.max(img) * 255)
           # else:
           #     img = img - self.displayMin
           #     img = (img / self.displayMax * 255)
           
           img = img.astype('uint8')

           imgZoom = self.zoom(img)
       
           self.displayImage = np.empty((imgZoom.shape[0], imgZoom.shape[1], 4), np.uint8, 'C')       
           self.displayImage[...,0] = imgZoom[...,2]
           self.displayImage[...,1] = imgZoom[...,1]
           self.displayImage[...,2] = imgZoom[...,0]
           self.displayImage[...,3].fill(255)
    
           self.displayImage2 = self.displayImage.copy()
           self.image = QtGui.QImage(self.displayImage2.copy(), self.displayImage.shape[1], self.displayImage.shape[0], QtGui.QImage.Format_RGB32)
           scaledSize = QtCore.QSize(self.geometry().width(), self.geometry().height()-40)

           self.pmap = QtGui.QPixmap.fromImage(self.image).scaled(scaledSize, QtCore.Qt.KeepAspectRatio)
    
           self.setPixmap(self.pmap)
       
       
       
   def zoom(self, img):
        """ Returns a zoomed version of the image for display.
        """

        if self.isZoomEnabled and self.zoomLevel > 0:               
            
            scaleFactor = 2**self.zoomLevel
            
            shape = np.array(img.shape)[0:2]
          
            # If we showing at least four pixels, calculate zoom
            if max(np.round(shape / (2**self.zoomLevel))) < 4:
                self.zoomLevel = round( np.log2(max(shape)/4) * self.zoomStepDivider  ) / self.zoomStepDivider
            
            zoomShape = np.round(shape / (2**self.zoomLevel))
            
            imageAspectRatio = np.shape(img)[1]/np.shape(img)[0]
            aspectRatio = self.geometry().width() / self.geometry().height() 
            
            
            if imageAspectRatio > aspectRatio:
                self.displayW = int(zoomShape[1]) 
                self.displayH =  min(shape[0], int(round(self.displayW / aspectRatio)))
     
            else:
                self.displayH = int(zoomShape[0]) 
                self.displayW = min(shape[1], int(round(self.displayH * aspectRatio))) 

            # Sets the current view position within the image based on the zoom
            # level and our scrolling position
            if self.zoomLocation is not None:
                h,w = self.currentImage.shape[0:2]
                                
                self.displayX = int(round(self.zoomLocation[0] - zoomShape[0] / 2))
                self.displayY = int(round(self.zoomLocation[1] - zoomShape[1] / 2))
                self.displayX = min(max(self.displayX ,0), w - self.displayW)
                self.displayY = min(max(self.displayY ,0), h - self.displayH)
               
                self.zoomLocation = None                          
           
            # Crop the image so we show only the zoomed area 
            if self.imageMode == self.RGB:
                imgCrop = img[self.displayY: self.displayY + self.displayH, self.displayX: self.displayX + self.displayW,:]
            else:
                imgCrop = img[self.displayY: self.displayY + self.displayH, self.displayX: self.displayX + self.displayW]

            zoomImage = imgCrop
            
            
        
        else:  #no Zoom
          
            self.displayX = 0
            self.displayY = 0
            self.displayW = img.shape[1]
            self.displayH = img.shape[0]
            zoomImage = img    
               
        return zoomImage
    
   
   def resizeEvent(self, new):
            """ Redraw if resized """
            self.set_image(self.currentImage)
       
      
   def screen_coords(self, x, y):
       """ Convert image co-ordinates to screen co-ordinates """
       if self.currentImage is None or self.pmap is None:
           return None, None
       else:
           xOffset, yOffset = self.screen_offsets()
           screenX = round((x - self.displayX) * (self.pmap.width()) / np.shape(self.displayImage)[1] + xOffset)
           screenY = round((y - self.displayY) * (self.pmap.height()) / np.shape(self.displayImage)[0] + yOffset)
           return screenX, screenY
       
       
   def image_coords(self, x, y):
       """ Convert screen co-ordinates to floor image co-ordinates """    

       if self.currentImage is None or self.pmap is None:
           return None, None
       else:
           xOffset, yOffset = self.screen_offsets()
           imageX = math.floor( (x - xOffset) / (self.pmap.width()) * np.shape(self.displayImage)[1] + self.displayX)
           imageY = math.floor( (y - yOffset) / (self.pmap.height()) * np.shape(self.displayImage)[0] + self.displayY)
           return imageX, imageY
   
        
   def image_coords_closest(self, x, y):
       """ Convert screen co-ordinates to cloest image co-ordinates """    

       if self.currentImage is None or self.pmap is None:
           return None, None
       else:
           xOffset, yOffset = self.screen_offsets()
           imageX = round( (x - xOffset) / (self.pmap.width()) * np.shape(self.displayImage)[1] + self.displayX)
           imageY = round( (y - yOffset) / (self.pmap.height()) * np.shape(self.displayImage)[0] + self.displayY)
           return int(imageX), int(imageY)
       
  
   def screen_offsets(self):  
           """ Returns the x and y co-ordinates of the top left of the image relative to Widget"""
           xOffset = (self.width() - self.pmap.width())/ 2 
           yOffset = (self.height() - self.pmap.height())/ 2 
           return int(xOffset), int(yOffset)
       
       
   def screen_size(self):  
           """ Returns the width and height of the view window in screen pixels"""
           return self.pmap.width(), self.pmap.height()
       
           
   def screen_dims(self, x,y):       
           """ Convert image dimensions to screen dimensions """
           screenX = round(x * (self.pmap.width()) / np.shape(self.displayImage)[1])
           screenY = round(y * (self.pmap.height()) / np.shape(self.displayImage)[0])               
           return int(screenX), int(screenY)
       
        
   def paintEvent(self, event):
       """ Handles drawing of widget, including labels, overlays etc.     """
       

       super().paintEvent(event)
       self.draw()


   def draw(self):
       """ This is where the whole thing is drawn"""       

       self.lastFrameTime = time.perf_counter()
       painter = QPainter(self)
       
       # Prevent drawing outside of image
       painter.setClipRect(QRect(*self.screen_offsets(), *self.screen_size() ))

       if self.imageMode == self.GRAPH:

           if self.graphData is not None:
               num_points = np.shape(self.graphData)[0]  
                 
               max_val = np.max(self.graphData)
               min_val = np.min(self.graphData)
               
               if self.graphDisplayMin is None:
                   graphMin = min_val
               else:
                   graphMin = self.graphDisplayMin
               
               if self.graphDisplayMax is None:
                   graphMax = max_val
               else:
                   graphMax = self.graphDisplayMax#
               plotX = np.linspace(0, self.graph_width, num_points)
               plotY = self.graph_height - (self.graphData - graphMin) / ( graphMax - graphMin) * self.graph_height
              
               painter.setPen(self.graphPen)

               for i in range(num_points - 1):
               
                   x,y = self.screen_coords(plotX[i],plotY[i])
                   x2,y2 = self.screen_coords(plotX[i + 1],plotY[i+1])
                   painter.drawLine(x,y,x2,y2)
                   
           font = painter.font()
           fm = QFontMetrics(font)        

           painter.setPen(self.graphLabelPen)
           prec = 3
           
           
           # Max val
           x,y = self.screen_coords(0,0)
           painter.drawText(x + 4, y + fm.height(), str(round(graphMax,prec)))
           painter.drawLine(x,y,x + 20,y) 
           
           # Min Val
           x,y = self.screen_coords(0,self.graph_height)
           painter.drawText(x + 4, y - fm.height(), str(round(graphMin,prec)))
           painter.drawLine(x,y,x + 20,y) 
           
           
           # Zero Line
           if graphMax > 0 and graphMin < 0:

               painter.setPen(self.graphZeroPen)

               x,y = self.screen_coords(0, self.graph_height - (0 - graphMin) / ( graphMax - graphMin) * self.graph_height)
               painter.drawLine(x,y,x + 100000,y) 
               
  

            
       #################### Draw overlays
       for overlay in self.overlays:
           
           painter.setPen(overlay.pen)
           if overlay.fill is not None:
               painter.setBrush(overlay.fill)
           
           # Convert the image co-ordinates to screen co-ordinates    
           x,y = self.screen_coords(overlay.x1, overlay.y1)
           w,h = self.screen_dims(overlay.x2, overlay.y2)
           
           if overlay.overlayType == self.ELLIPSE:    
               painter.drawEllipse(x,y,w,h)
           elif overlay.overlayType == self.RECTANGLE:
               painter.drawRect(x,y,w,h)
           elif overlay.overlayType == self.POINT:
               painter.drawPoint(x,y)
           elif overlay.overlayType == self.LINE:
               painter.drawLine(x,y,x + w,y + h)
           elif overlay.overlayType == self.TEXT:
               painter.drawText(x, y, overlay.text)
               
               
       ##################### Draw dragging ROI    
       if self.dragging and self.dragX != self.dragToX and self.dragY != self.dragToY:
           painter.setPen(QPen(Qt.red, 2, Qt.SolidLine))
          
           dragRoi = int(round(min(self.dragX, self.dragToX))), int(round(min(self.dragY, self.dragToY))), int(round(max(self.dragX, self.dragToX))), int(round(max(self.dragY, self.dragToY)))

           drawX, drawY = self.screen_coords(dragRoi[0], dragRoi[1])
           endX, endY = self.screen_coords(dragRoi[2], dragRoi[3])
           width = endX- drawX
           height = endY - drawY

           painter.setPen(self.roiDragContrastPen)
           painter.drawRect(drawX, drawY, width + 1, height + 1)
           painter.setPen(self.roiDragPen)
           painter.drawRect(drawX, drawY, width + 1, height + 1)
           
                
           
           
       ##################### Draw active ROI 
       if self.roi is not None:
           drawX, drawY = self.screen_coords(self.roi[0], self.roi[1])
           endX, endY = self.screen_coords(self.roi[2], self.roi[3])
           width = endX- drawX
           height = endY - drawY
           painter.setPen(self.roiContrastPen)
           painter.drawRect(drawX, drawY, width + 1, height + 1)
           painter.setPen(self.roiPen)
           painter.drawRect(drawX, drawY, width + 1, height + 1)

          
       # Stop clipping because now we want to draw status bar which is outside image
       painter.setClipRect(QRect(*self.screen_offsets(), self.screen_size()[0],100000))
       
       
       ##################### Draw Zoom Indicator ############
       if self.zoomLevel > 0 and self.isZoomIndicator:

           painter.setPen(self.zoomIndicatorPen)
           
           if self.imageHeight < self.imageWidth:
               x = self.width() - (self.width() - self.pmap.width()) / 2 - self.zoomIndicatorOffsetX - self.zoomIndicatorWidth
               y = (self.height() - self.pmap.height()) / 2 + self.zoomIndicatorOffsetY
               w = self.zoomIndicatorWidth
               h = self.imageSize[0] / self.imageSize[1] * w
           else:
               x = self.width() - (self.width() - self.pmap.width()) / 2 - self.zoomIndicatorOffsetX 
               y = (self.height() - self.pmap.height()) / 2 + self.zoomIndicatorOffsetY
               h = self.zoomIndicatorWidth
               w = self.imageSize[1] / self.imageSize[0] * h
               
           # Outer rect
           painter.drawRect(int(x), int(y), int(w), int(h)) 
           
           # Filled inner rect
           x2 = round(self.displayX / self.imageSize[1] * (w- 2))
           y2 = round(self.displayY / self.imageSize[0] * (h - 2))
           w2 = round(self.displayW / self.imageSize[1] * (w - 2) + 1)
           h2 = round(self.displayH / self.imageSize[0] * (h - 2) + 1)
           w2 = min(w - x2, w2 - x2) + x2
           h2 = min(h - y2, h2 - y2) + y2
         
           painter.setBrush(self.zoomIndicatorBrush)
           painter.drawRect(int(x + x2), int(y + y2), int(w2), int(h2))  

      
       ##################### Draw status bar 
       if self.isStatusBar and self.pmap is not None:
       
           font = painter.font()
           fm = QFontMetrics(font)
           if self.mouseX is not None and self.mouseY is not None and self.currentImage is not None:
               try:
                   mX = str(round(self.mouseX))
                   mY = str(round(self.mouseY))
                   if self.imageMode == self.MONO:
                       cursorVal = str(int(round(self.currentImage[round(self.mouseY), round(self.mouseX)],1)))
                   elif self.imageMode == self.RGB:
                       cursorVal = str(int(round(self.currentImage[round(self.mouseY), round(self.mouseX),0],0))) \
                                   + "," + str(int(round(self.currentImage[round(self.mouseY), round(self.mouseX),1],0))) \
                                   + "," + str(int(round(self.currentImage[round(self.mouseY), round(self.mouseX),2],0)))
                  
                   else:
                       cursorVal = '--'
               except:
                   mX = '-'
                   mY = '-'
                   cursorVal = '--'
           else:
               mX = '-'
               mY = '-'
               cursorVal = '--'
                         
           if self.currentImage is not None:
               self.meanPixel = str(round(np.mean(self.currentImage),1))
               self.maxPixel = str(round(np.max(self.currentImage)))
               self.minPixel = str(round(np.min(self.currentImage)))
           else:
               self.meanPixel = '-'
               self.maxPixel = '-'
               self.minPixel = '-'
               
           if self.currentImage is not None and self.roi is not None and self.roi is not []:
               roi = self.currentImage[self.roi[1] : self.roi[3], self.roi[0]: self.roi[2],...]
               self.roiMax = str(round(np.max(roi)))
               self.roiMin = str(round(np.min(roi)))
               self.roiMean = str(round(np.mean(roi),1))               
           
           if self.zoomLevel > 0:
               text = str(round(2**(self.zoomLevel))) + 'X '
           else:
               text = ''
               
           if self.imageMode is not self.GRAPH:    
               text = text + '(' + mX + ',' + mY + ') = ' + cursorVal + ' | [' + self.minPixel + '-' + self.maxPixel + ', Mean: ' + self.meanPixel + ']'
           
               if self.roi is not None:
                   text = text + ' | [ROI: (' + str(self.roi[0] ) + ',' + str(self.roi[1]) + ')-(' + str(self.roi[2] -1) + '-' + str(self.roi[3] -1) + '): ' + self.roiMin + '-' + self.roiMax + ', Mean: ' + self.roiMean + ']' 
            
               elif self.dragging and self.dragX != self.dragToX and self.dragY != self.dragToY:
                   text = text + ' | [Dragging ROI: (' + str(dragRoi[0]) + ',' + str(dragRoi[1] ) + ')-(' + str(dragRoi[2] -1) + '-' + str(dragRoi[3] -1) + ') ]'
           else: #GRAPH
               if self.mouseX is not None and self.mouseY is not None:
                   xP =  str(round(int(self.mouseX / self.graph_width * num_points) ))
                   yP =  str(round(self.graphData[int(self.mouseX / self.graph_width * num_points) ],3))
    
                   prec = 3
                   text = text + '(' + xP + ',' + yP + ') | [' + str(round(np.min(self.graphData),prec)) + ' -> ' + str(round(np.max(self.graphData),prec)) + ', Mean: ' + str(round(np.mean(self.graphData),prec)) + ']'
                    
                   x,y = self.screen_coords(self.mouseX,plotY[int(self.mouseX* num_points / self.graph_width)])
                   painter.setBrush(self.graphCursorBrush)
                   painter.setPen(self.graphCursorPen)
                   painter.drawEllipse(x-self.graphCursorSize,y -self.graphCursorSize, self.graphCursorSize * 2,self.graphCursorSize *2)

           xPos = int(round((self.width() - self.pmap.width()) /2))
           painter.setBrush(self.statusBrush)
           painter.setPen(self.statusPen)
           painter.drawRect(int(xPos + 1), int(self.height() - fm.height() - 5), int(self.pmap.width() - 2), int(fm.height() + 8))
           painter.setPen(self.statusTextPen)
           painter.drawText(int(xPos + 10), int(self.height() - 6), text)
          
       self.update() 
        
   def add_overlay(self, overlayType, *args):
       """ Adds an overlay to the list of visible overlays. overlayType can be ELLIPSE, RECTANGLE, POINT or LINE.
       add_overlay(ImageDisplay.ELLIPSE, x, y, width, height, pen, fill)
       add_overlay(ImageDisplay.RECTANGLE, x, y, width, height, pen, fill)
       add_overlay(ImageDisplay.LINE, x, y, width, height, pen, fill)
       add_overlay(ImageDisplay.POINT, x, y, pen, fill)
       add_overlay(ImageDisplay.TEXT, x, y, pen, text)       
       """       
       x = args[0]
       y = args[1]
       text = ''
       
       if overlayType == self.ELLIPSE or overlayType == self.RECTANGLE:
           w = args[2]
           h = args[3]
           pen = args[4]
           fill = args[5]
    
       elif overlayType == self.LINE:
           w = args[2]
           h = args[3]
           pen = args[4]
           fill = None
                    
       elif overlayType == self.POINT:
           w = 1
           h = 1
           pen = args[2]
           fill = None
      
           
       elif overlayType == self.TEXT:
           w = 1
           h = 1
           pen = args[2]
           fill = None
           text = args[3]
           
       else:
           overlayType = None
       
       if overlayType is not None:
           newOverlay = Overlay(overlayType, x, y, w, h, pen, fill, text)
           self.overlays.append(newOverlay)
           return newOverlay
       else:
           return None
       
   
   def remove_overlay(self, overlay):
       """ Removes overlay 'overlay' from the list of visible overlays
       """
       self.overlays.remove(overlay)
       self.update()
         
           
   def clear_overlays(self):
       """ Removes all overlays from the list of visible overlays
       """
       self.overlays = []
       self.update()
           
   
   def num_overlays(self):
       """ Returns the number of overlays in the list of visible overlays
       """
       return len(self.overlays)
   
    
   def set_auto_scale(self, autoScale):
       """ Determines whether or not images are brightness/contrast autoscaled to use the full dynamic range. True or False.
       """
       self.autoScale = autoScale
       self.update()
       
       
   def set_status_bar(self, isStatusBar):
       """ Set status bar visible (True) or hidden (False).
       """
       self.isStatusBar = isStatusBar
       self.update()
       
    
   def set_zoom_enabled(self, isZoomEnabled):
       """ Set whether zoom using mouse wheel is possible (True) or not (False).
       """
       self.isZoomEnabled = isZoomEnabled
       if not isZoomEnabled:
           self.zoomLevel = 0
       self.set_image(self.currentImage)
       self.update()

   def set_zoom_indicator_enabled(self, isIndicatorEnabled):
       """ Set whether the zoom indicator is shown (True) or not (False).
       """
       self.isZoomIndicator = isIndicatorEnabled
       self.set_image(self.currentImage)
       self.update()
       
    
   def set_roi_enabled(self, isRoiEnabled):
       """ Set whether zoom using mouse wheel is possible (True) or not (False).
       """
       self.isRoiEnabled = isRoiEnabled
       if not self.isRoiEnabled:
           self.roi = None
       self.update()
         
    
   def set_colormap(self, colormapName):
       """ Sets the colormap using a matplotlib colormap. Provide the colormap as a string containing the anme of the colormap.
       """

       if colormapName is None:
           self.colortable = None
       else:   
           colormap = cm.get_cmap(colormapName) 
           colormap._init()
           lut = (colormap._lut * 255).view(np.ndarray).astype('int16')  # Convert matplotlib colormap from 0-1 to 0 -255 for Qt
           lut = lut[:256,0:3]
           nCols = np.shape(lut)[0]   # Seem to be 513 x 4
               
           self.colortable = []
           for i in range(0, 256):
               col = int(round(i * nCols / 256))
               self.colortable.append(QtGui.qRgb(lut[col,0],lut[col,1],lut[col,2]))
       self.update()
       
       
   def set_display_range(self, lower, upper):
       """ Sets the intensity range used for display if set_auto_scale is False. Pixels
       of 'lower' or below will be mapped to 0, 'upper' and above to 255.
       """
       self.displayMin = lower
       self.displayMax = upper
       self.update()
       

   def get_display_range(self):
       """ Gets the intensity range used for display if set_auto_scale is False. Return
       a tuple of (lower, upper).
       Pixels of 'lower' or below will be mapped to 0, 'upper' and above to 255.
       """
       return self.displayMin, self.displayMax
   
       
   def set_zoom_step_divider(self, zoomDivider):
       """ Sets the zoom step divider. A bigger value makes the amount zoomed
       in by each mouse wheel step smaller.
       """
       if zoomDivider > 0:       
           self.zoomStepDivider = zoomDivider
           
         
class Overlay():
    """ Class to store details about an overlay.
    """
    
    x1 = None  
    x2 = None
    y1 = None
    y2 = None
    pen = None
    fill = None
    overlayType = None
    text = None
    
    def __init__(self, overlayType, x1, y1, x2, y2, pen, fill, text):

        self.overlayType = overlayType
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2
        self.pen = pen
        self.fill = fill
        self.text = text


# -*- coding: utf-8 -*-
"""
Created on Tue Mar 15 16:12:22 2022

@author: AOG
"""


from PyQt5 import QtGui, QtCore, QtWidgets
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from PyQt5.QtWidgets import *
from PyQt5.QtGui import QPalette, QColor, QImage, QPixmap, QPainter, QPen, QGuiApplication
from PyQt5.QtGui import QPainter, QBrush, QPen



class DoubleSlider(QSlider):

    # create our our signal that we can connect to if necessary
    doubleValueChanged = pyqtSignal(float)

    def __init__(self, decimals=3, *args, **kargs):
        super(DoubleSlider, self).__init__( QtCore.Qt.Horizontal) #*args, **kargs)
        self._multi = 10 ** decimals

        self.valueChanged.connect(self.emitDoubleValueChanged)

    def emitDoubleValueChanged(self):
        value = float(super(DoubleSlider, self).value())/self._multi
        self.doubleValueChanged.emit(value)

    def value(self):
        return float(super(DoubleSlider, self).value()) / self._multi

    def setMinimum(self, value):
        return super(DoubleSlider, self).setMinimum(value * self._multi)

    def setMaximum(self, value):
        return super(DoubleSlider, self).setMaximum(value * self._multi)

    def setSingleStep(self, value):
        return super(DoubleSlider, self).setSingleStep(value * self._multi)

    def singleStep(self):
        return float(super(DoubleSlider, self).singleStep()) / self._multi

    def setValue(self, value):
        super(DoubleSlider, self).setValue(int(value * self._multi))    



# -*- coding: utf-8 -*-
"""
CAS: Camera Acquisition System

A camera interface that extends WebCamera to return colour images.

"""

import cv2 as cv
import time

import numpy as np
from cas_gui.cameras.WebCamera import WebCamera  

        
class WebCameraColour(WebCamera):

          
    def grab_image(self):
        
        rval, imageData = self.vc.read()
        imageData = cv.cvtColor(imageData, cv.COLOR_BGR2RGB)
        
        return imageData.astype('uint8')


    def is_colour(self):
        return True

if __name__ == "__main__":
    print("Test mode not implemented")


# -*- coding: utf-8 -*-
"""
#  HughesLab: im_tools

"""

import os
import math

import numpy as np
import numpy.ma as ma

from PIL import Image



def load_image(filename):
    """
    Loads an image or stack of images from a file. 
    
    Arguments:
        filename    : str or Path
                      path to file
    
    Returns:
        ndarray     : 2D, 3D or 4D numpy array representing image or stack  

    """
    im = Image.open(filename)
    
    if im.n_frames > 1:
        h,w = np.shape(im)
        dt = np.array(im).dtype
        stack = np.zeros((im.n_frames, h,w), dtype = dt)
        for i in range(im.n_frames):
            im.seek(i)
            stack[i,:,:] = np.array(im)
        return stack
    else:    
        return np.array(Image.open(filename))


def crop_zero(img):
    """ 
    Crops an image to the smallest rectangle that contains all non-zero
    pixels.
    
    Returns cropped image as 2D numpy array.
    
    Arguments:
        img      : ndarray
                   image as numpy array
                   
    Returns
        ndarray, cropped image                
    """
    
    hMax = np.amax(img, axis = 0)

    hCrop = (np.argwhere(hMax > 0)[0].item(), np.argwhere(hMax > 0)[-1].item() )
    
    vMax = np.amax(img, axis = 1)
    vCrop = (np.argwhere(vMax > 0)[0].item(), np.argwhere(vMax > 0)[-1].item() )
    
    img = img[vCrop[0]:vCrop[1],hCrop[0]:hCrop[1]]
    
    return img
    
    

def extract_central(img, boxSize = None):
    """ 
    Extract a central square from an image. The extracted square is centred
    on the input image, with size 2 * boxSize if possible, otherwise the largest
    square that can be extracted.
    
    Returns cropped image as 2D numpy array.
    
    Arguments:
        img     : input image as 2D numpy array
        
    Keyword Arguments:    
        boxSize : size of cropping square, default is largest possible
        
    Returns:
        ndarray, cropped image
    """

    w = np.shape(img)[0]
    h = np.shape(img)[1]

    if boxSize is None:
        boxSize = min(w,h)
    cx = w/2
    cy = h/2
    boxSemiSize = min(cx,cy,boxSize)
    
    imgOut = img[math.floor(cx - boxSemiSize):math.floor(cx + boxSemiSize), math.ceil(cy- boxSemiSize): math.ceil(cy + boxSemiSize)]
    
    return imgOut


def to8bit(img, minVal = None, maxVal = None):
    """ Returns an 8 bit representation of image. If min and max are specified,
    these pixel values in the original image are mapped to 0 and 255 
    respectively, otherwise the smallest and largest values in the 
    whole image are mapped to 0 and 255, respectively.
    
    Arguments:
            img    : ndarray
                     input image as 2D numpy array
        
    Keyword Arguments:    
            minVal : float
                     optional, pixel value to scale to 0
            maxVal : float
                     optional, pixel value to scale to 255
    """
 
    img = img.astype('float64')
       
    if minVal is None:
        minVal = np.min(img)
                    
    img = img - minVal
        
    if maxVal is None:
        maxVal = np.max(img)
    else:
        maxVal = maxVal - minVal
        
    img = img / maxVal * 255
    img = img.astype('uint8')
    
    return img


def to16bit(img, minVal = None, maxVal = None):
    """ Returns an 16 bit representation of image. If min and max are specified,
    these pixel values in the original image are mapped to 0 and 2^16 
    respectively, otherwise the smallest and largest values in the 
    whole image are mapped to 0 and 2^16 - 1, respectively.
    
    Arguments:
        img    : ndarray
                 input image as 2D numpy array
        
    Keyword Arguments:    
        minVal : float
                 optional, pixel value to scale to 0
        maxVal : float
                 optional, pixel value to scale to 2^16 - 1
                 
    Returns:
        ndarray, 16 bit image             
    """   
        
    img = img.astype('float64')
       
    if minVal is None:
        minVal = np.min(img)
                    
    img = img - minVal
        
    if maxVal is None:
        maxVal = np.max(img)
    else:
        maxVal = maxVal - minVal
        
    img = img / maxVal * (2^16 - 1)
    img = img.astype('uint16')
    
    return img


def radial_profile(img, centre):
    """ Produce angular averaged radial profile through image img centred on
    centre, a tuple of (x_centre, y_centre)
    
    Returns radial profile as 1D numpy array

    Arguments:
        img    : ndarray
                 input image as 2D numpy array
        centre : (int, int)
                 centre point for radial profile, tuple of (x,y)  
    Returns:
        ndarray, 1D profile             
    """
    
    y, x = np.indices((img.shape))
    r = np.sqrt((x - centre[1])**2 + (y - centre[0])**2)
    r = r.astype(int)
    
    tbin = np.bincount(r.ravel(), weights = img.ravel())
    nr = np.bincount(r.ravel())
    radialprofile = tbin  / nr
    
    return radialprofile 
 



def save_image8(img, filename):
    """ Saves image as 8 bit tif without scaling.
    
    Arguments:
         img      : ndarray, 
                    input image as 2D numpy array
                   
         filename : str
                    path to save to, folder must exist
    """
    
    im = Image.fromarray(img.astype('uint8'))
    im.save(filename)



def save_image16(img, filename):
    """ Saves image as 16 bit tif without scaling.
        
    Arguments:
         img      : ndarray, 
                    input image as 2D numpy array
                   
         filename : str
                    path to save to, folder must exist
    """
    im = Image.fromarray(img.astype('uint16'))
    im.save(filename)


     
def save_image8_scaled(img, filename):
    """ Saves image as 8 bit tif with scaling to use full dynamic range.
            
    Arguments:
         img      : ndarray, 
                    input image as 2D numpy array
                   
         filename : str
                    path to save to, folder must exist
    """
    
    im = Image.fromarray(to8bit(img))
    im.save(filename)
    
    
def save_image16_scaled(img, filename):
    """ Saves image as 16 bit tif with scaling to use full dynamic range.
            
    Arguments:
         img      : ndarray, 
                    input image as 2D numpy array
                   
         filename : str
                    path to save to, folder must exist
    """
        
    im = Image.fromarray(to16bit(img)[0])
    im.save(filename) 


def average_channels(img):
    """ Returns an image which is the the average pixel value across all channels of a colour image.
    It is safe to pass a 2D array which will be returned unchanged.
    
    
    Arguments:
        img:    ndarray
                 image as 2D/3D numpy array
            
    Returns:
        ndarray, averaged image        
    """     

    if img.ndim == 3:
        return np.mean(img, 2)
    else:
        return img
    
    
def max_channels(img):
    """ Returns an image which is the the maximum pixel value across all channels of a colour image.
    It is safe to pass a 2D array which will be returned unchanged.
    
    
    Arguments:
        img:   ndarray
               image as 2D/3D numpy array
               
    Returns:
        ndarray, max value image           
    """     

    if img.ndim == 3:
        return np.max(img, 2)
    else:
        return img
        
        
    

def load_stack(folder):
    """ Loads a stack of images from a folder into a 3D numpy array
    
    Arguments:
        folder      : str
                      path to folder
        status      : boolean
                      if True, updates status on console (default = True)              
                      
    Returns:
        ndarray, 3D numpy array (im, y, x)                     
    """    

    image_files = [f for f in os.listdir(folder)]
    
    nImages = len(image_files)    
    
    testIm = np.array(Image.open(os.path.join(folder,image_files[0])))
    
    h, w = np.shape(testIm)    
    
    data = np.zeros((nImages, h, w), dtype = testIm.dtype)
   
   
    
    for idx, image_file in enumerate(image_files):            
        data[idx,:,:] = np.array(Image.open(os.path.join(folder,image_file)))        
        
    return data   
  

def save_tif_stack(stack, filename, bit_depth = 8, auto_contrast = None, fixed_min = None):
    """ Writes stack of images from 3D numpy array to file. The array must 
    be orders (frame, y, x).
    
    Arguments:
        stack         : ndarray
                        3D numpy array (frame, y, x)
        filename      : str or Path
                        path to file name. Folder must exist.            
                       
    Keyword Arguments:
        bit_depth     : int
                        8 (default) or 16
        auto_contrast : str or None
                        Whether or not to scale images to use full bit depth
                        'image' to autoscale each image individually
                        'stack' to autoscale entire stack
                        None or 'none' for no autoscaling
        fixed_min     : int or None
                        if auto_contrast is 'image' or 'stack', setting this
                        value fixed the lower range of the saved image pixel
                        values rather than taking the minimum from the images.
                                         
    """
    if stack.ndim != 3:
       raise Exception("Stack must be 3D array.")

    if auto_contrast == 'stack':
       maxVal = np.max(stack)
       if fixed_min is None:
           minVal = np.min(stack)
       else:
           minVal = fixed_min
    elif auto_contrast == 'image' or auto_contrast == 'none' or auto_contrast is None:
        pass
    else:
        raise Exception("Keyword auto_contrast only accepts 'stack', 'image' or None.")

       
    if bit_depth == 16:
        dt = 'uint16'
    elif bit_depth == 8:
        dt = 'uint8'
    else:
        raise Exception("Bit depth can only be 8 or 16.")
    
    imlist = []
    for im in stack:
        
        if auto_contrast == 'image':
            maxVal = np.max(np.abs(im))
            if fixed_min is None:
                minVal = np.min(np.abs(im))
            else:
                minVal = fixed_min
                
        if auto_contrast is not None:
            im = im.astype('float64') - minVal
            intrange = maxVal - minVal
            if intrange > 0:
                im = im / intrange * (2**bit_depth - 1)
            else:
                im[:] = 2**bit_depth - 1
        imlist.append(Image.fromarray(im.astype(dt)))


    imlist[0].save(filename, compression="tiff_lzw", save_all=True,
           append_images=imlist[1:])
    
    


def log_scale_image(img, min_val = None):
    """ Generates a log-scaled image, adjusted for good visual appearance,
    particularly for OCT images.
    
    Arguments:
        img     : ndarray 
                  image as 2D numpy array
     
    Optional Keyword Arguments:
        min_val : float
                  if specified, 0 in image will correspond to this value 
                  in log scaled image before windowing
                  
    Returns:
        ndarray, log-scaled image (8 bit)             
                   .
    """

    if min_val is None:
        dispMin = log_scale_min(img)
        
    else:
        dispMin = min_val
    
    imgOut = np.log(img + 0.00001)
    imgOut = imgOut - dispMin
    maxVal = np.max(imgOut) * 0.9
    imgOut = imgOut / maxVal
    imgOut[imgOut < 0] = 0
    imgOut[imgOut > 1] = 1
    
    return imgOut


def log_scale_min(img):
    """Calculates minimum display value for a log-scaled image so that 
    it does no appear overly noisy.
    
    Arguments:
        img     : ndarray 
                  image as 2D numpy array
     
    
    Returns:
        float, minimum value           
                   .
    """    
    
    mask = img == 0
    im = np.log(img + 0.00001)
    imgOut = ma.masked_array(im, mask)
    valRange = np.max(imgOut) - ma.min(imgOut)

    lineAv = ma.mean(imgOut,0)
    sVals = np.sort(lineAv)

    minVal = sVals[min(10, len(sVals) - 1)]

    dispMin = minVal - valRange / 25
    if dispMin is not ma.masked:
        return dispMin
    else:
        return 0    


# -*- coding: utf-8 -*-
"""
ImageProcessorThread
Part of CAS GUI: Camera Acquisition System GUI

Threading class for image processing for use in GUIs or or 
multi-threading applications. Sub-class and implement process_frame
to make a custom image processor.

"""

import queue
import threading
import time
import logging
import multiprocessing
import numpy as np

from cas_gui.threads.image_processor_process import ImageProcessorProcess


class ImageProcessorThread(threading.Thread):
    
    process = None
    updateQueue = None
    sharedMemory = None
    sharedMemoryArray = None
    inputSharedMemory = None
    imSize = (0,0)
    lastImNum = -1  
    numDropped = 0
    
    def __init__(self, processor, inBufferSize, outBufferSize, **kwargs):
        
        # Caller passes the name of the camera class (which should be in a
        # module of the same name) in the variable camName. This is dynamically
        # imported here and an instance created as self.cam.
   
        super().__init__()
        
        self.processor = processor()
        self.inBufferSize = inBufferSize
        self.outBufferSize = outBufferSize
        self.inputQueue = kwargs.get('inputQueue', None)
        self.acquisitionLock = kwargs.get('acquisitionLock', None)
        self.multiCore = kwargs.get('multiCore', False)
        self.useSharedMemory = kwargs.get('sharedMemory', False)        
        self.sharedMemoryArraySize = kwargs.get('sharedMemoryArraySize', (2048, 2048))

        if self.multiCore and not self.useSharedMemory:                
            self.outputQueue = multiprocessing.Queue(maxsize=self.outBufferSize)
        else:
            self.outputQueue = queue.Queue(maxsize=self.outBufferSize)
            
        self.statusQueue = multiprocessing.Queue(maxsize = 1000)    

        self.currentOutputImage = None
        self.currentInputImage = None
        
        self.lastFrameTime = 0
        self.frameStepTime = 0
        self.currentFrame = None
        self.currentFrameNumber = 0
        self.isPaused = False
        self.isStarted = True
        self.batchProcessNum = 1
        
        # If we are going to use a different core to do processing, then we need
        # to start a process on that core.
        
        if self.multiCore:
            
            # Queue for sending updates on how to do the processing
            self.updateQueue = multiprocessing.Queue()
            self.messageQueue = multiprocessing.Queue()
            
            if self.useSharedMemory:
                outQueue = None
            else:
                outQueue = self.outputQueue
                
            self.shareName = str(np.random.randint(6400000, size = 1)[0])
            
            # Create the process and set it running
            self.process = ImageProcessorProcess(self.inputQueue, outQueue, self.updateQueue, 
                                                 self.messageQueue, sharedMemoryArraySize = self.sharedMemoryArraySize, 
                                                 useSharedMemory = self.useSharedMemory, statusQueue = self.statusQueue, shareName = self.shareName)
            self.process.start()
            time.sleep(0.1)
            self.updateQueue.put(self.processor)
           
        
    
        
    # This loop is run once the thread starts
    def run(self):

         while self.isStarted:                 

            
             if self.multiCore:
                                   
                 if self.useSharedMemory:
                 
                     # If we have not yet got a reference to the shared memory, get it now
                     if self.sharedMemory is None:
                         temp = np.ndarray(self.sharedMemoryArraySize, dtype = 'float32')
                         try:
                             self.sharedMemory = multiprocessing.shared_memory.SharedMemory(create=True, size=temp.nbytes, name = self.shareName)
                         except:
                             self.sharedMemory = multiprocessing.shared_memory.SharedMemory(size=temp.nbytes, name = self.shareName)

                         #try:
                             #self.sharedMemory = multiprocessing.shared_memory.SharedMemory(name="CASShare")
                         self.sharedMemoryArray = np.ndarray(self.sharedMemoryArraySize, dtype = 'float32', buffer = self.sharedMemory.buf)  
                         #except:
                          #   pass
                     else:
                         # The image width, height, processing step time, and frame number are
                         # stored in a corner of the shared memory
                         imW = int(self.sharedMemoryArray[0,1])
                         imH = int(self.sharedMemoryArray[0,0])                       
                        
                         imNum = int(self.sharedMemoryArray[0,3])
                         imIndex = int(self.sharedMemoryArray[0,4])

                         if imNum > self.lastImNum + 1:
                             self.numDropped += 1
                             
                             
                         # Only pull off and return an image if we haven't already returned this
                         if imNum > self.lastImNum  and imW > 0 and imH > 0:
                             im = self.sharedMemoryArray[1:1+imH, :imW] #.copy()
                             #m = np.zeros((10,10))
                             self.lastImNum = imNum           
                         
                             if self.outputQueue.full():
                                 temp = self.outputQueue.get()
                             
                             self.outputQueue.put_nowait((im, imIndex)) 
                             #print("got out image")
                         else:
                              time.sleep(0.01)

                              
                 else:  # self.useSharedMemory = False
                      # If we are not using shared memory we don't need to do anything here because
                      # the processed images will be put straight in the output queue by the ImageProcessorProcess
                      time.sleep(0.01)
             
             else:    
                 
                 if not self.isPaused:
                     
                     self.handle_flags()   
                     
                     # Stop output queue overfilling
                     if self.outputQueue.full():
                         for i in range(self.batchProcessNum):
                             temp = self.outputQueue.get()
                   
                     if self.get_num_images_in_input_queue() >= self.batchProcessNum:                         
    
                         if self.acquisitionLock is not None: self.acquisitionLock.acquire()
                         
                         try:
                             if self.batchProcessNum > 1:
                                 
                                 img = self.inputQueue.get()
                                 
                                 self.currentInputImage = np.zeros((np.shape(img)[0], np.shape(img)[1], self.batchProcessNum))
                                 self.currentInputImage[:,:,0] = img
                                 
                                 for i in range(1, self.batchProcessNum):
                                     self.currentInputImage[:,:,i] = self.inputQueue.get()
                                 out = self.process_frame(self.currentInputImage)
                                 
                                 if isinstance(out, tuple):
                                     self.currentOutputImage = out[0]
                                 else:
                                     self.currentOutputImage = out
                                 
                                 self.outputQueue.put(self.currentOutputImage)
    
    
                             else:
                                 self.currentInputImage = self.inputQueue.get()
                                 self.currentOutputImage = self.process_frame(self.currentInputImage)
                                 self.outputQueue.put(self.currentOutputImage)
    
                             # Timing
                             self.currentFrameNumber = self.currentFrameNumber + 1
                             self.currentFrameTime = time.perf_counter()
                             self.frameStepTime = self.currentFrameTime - self.lastFrameTime
                             self.lastFrameTime = self.currentFrameTime
                     
                         except queue.Empty:
                             print("No image in queue")
                         
                         if self.acquisitionLock is not None: self.acquisitionLock.release()
    
                     else:
                         time.sleep(0.01)
                         
    def acquire_set(self):
    
        img = self.inputQueue.get()

        self.currentInputImage = np.zeros((np.shape(img)[0], np.shape(img)[1], self.batchProcessNum))
        self.currentInputImage[:,:,0] = img
        
        for i in range(1, self.batchProcessNum):
            self.currentInputImage[:,:,i] = self.inputQueue.get()                 
        
        return self.currentInputImage[:,:,:]    
         
                   
    def pipe_message(self, command, parameter):
        """ Sends a message to update an attribute or call a function in the
        ImageProcessorClass.
        
        Arguments:
            command   : str
                        name of attribute or function
            parameter : tuple
                        if calling a function, this is a tuple containing
                        the arguments to be passed to the function. If setting
                        an attribute, this is the value to set
        """
        if self.multiCore:
            self.messageQueue.put((command, parameter))
        else:
            if self.processor is not None:
                self.processor.message(command, parameter)
                

                
    def get_processor(self):
        """ Returns reference to processor object"""
        return self.processor
     
        

    def process_frame(self, inputFrame):
        """ Processes a single frame and return processed frame. Both are
        numpy arrays"""
        
        ret = self.processor.process(inputFrame) 
        return ret
        
    
    ######### Override with code that should be run on each iteration
    def handle_flags(self):
        pass
        
       
    def get_input_queue(self):
        return self.inputQueue
    
    
    def get_ouput_queue(self):
        return self.outputQueue
    
    
    def get_num_images_in_input_queue(self):
        return self.inputQueue.qsize()
    
    
    def get_num_images_in_output_queue(self):
        return self.outputQueue.qsize()
    
        
        
    
    def is_image_ready(self):
        """ Returns true if there is a new processed image ready to be read.
        """             
        if self.get_num_images_in_output_queue() > 0:
            return True
        else:
            return False
    
    
    def add_image(self, im):
        """ Adds a raw image to be processed.
        """
        # Stop output input overfilling
        if self.inputQueue.full():
            temp = self.inputQueue.get()

        self.inputQueue.put_nowait(im)
    
    
    
    def get_next_image(self):
        """ Returns the next available processed image.
        """        
        im = None         
        if self.is_image_ready() is True:   
            try:
                im = self.outputQueue.get_nowait()   
            except:
                im = None
        
        return im
            
        
    def get_actual_fps(self):
        """ Returns the processing frame rate.
        """
        
        # If using single core, we have already stored the frame step time
        if not self.multiCore and self.frameStepTime > 0:
            return (1 / self.frameStepTime)
        
        # If using shared memory we can pull this from the shared memory as the processor
        # stores it there
        elif self.multiCore and self.useSharedMemory and self.sharedMemoryArray is not None:
            if self.sharedMemoryArray[0,2] > 0:
                return (1 / self.sharedMemoryArray[0,2])
            else:
                return 0
        
        # If using multiCore without shared memory, we do not have access to frame rate, so
        # return 0
        else:
            return 0
        
    
    
    def get_latest_processed_image(self):
        """ Returns the last processed image that was obtained.
        """
        return self.currentOutputImage
    
   
    
    def get_latest_input_image(self):
        """ Returns the last raw image that was added.
        """
        return self.currentInputImage    
    
   
    
    def flush_input_buffer(self):
        """ Removes all raw images from queue.
        """
        with self.inputQueue.mutex:
            self.inputQueue.queue.clear()
            
    def flush_status_buffer(self):
        """ Removes all messages from status queue
        """
        while True:
            try:
                null = self.statusQueue.get_nowait()
            except:
                break
            
            
            
    def set_batch_process_num(self, num):
        """ Sets the size of the batch of images to be sent to the processor.
        """
        self.batchProcessNum = num
        if self.multiCore:
            self.pipe_message("set_batch_process_num", num)
        
    
    def flush_output_buffer(self):
        """ Removes all processed images from queue if not using shared memory
        """
        if self.useSharedMemory is False:
            with self.outputQueue.mutex:
                self.outputQueue.queue.clear()    
   
            
    def pause(self):
        """ Pauses the processing.
        """
        self.isPaused = True
        return
    

    def resume(self):
        """ Resumes paused processing.
        """
        self.isPaused = False
        return            
    
          
    def stop(self):
        """ Stops the process. 
        """
        self.isStarted = False
        if self.process is not None:
            print("Terminating process")
            self.process.terminate()
            self.process.join()

  
    def update_settings(self):
        """ Sends a copy of the processor class to the process running on
        another core. """
        if self.updateQueue is not None:
            self.updateQueue.put(self.processor)


# -*- coding: utf-8 -*-
"""
Kent-CAS: Camera Acquisition System

Threading class for image acqusition and buffering for use in GUIs or or 
multi-threading applications.

"""

import queue
import multiprocessing
from multiprocessing import shared_memory
import threading
import time
import importlib

class ImageAcquisitionThread(threading.Thread):
    """ Threading class to acquire images from camera.
    
    Arguments:
        camName    : str
                     name of camera class. The class must be in a module
                     of the same name, either in the current path or
                     in the cameras folder of CAS GUI
                     
    Keyword Arguments:    
        bufferSize : int
                     size of queue to store acquired images, default is 10
        aquisitionLock : 
        imageQueue : Queue or None
                     Provide a queue here if direct access is required, 
                     otherwise one will be created and used internally
        auxillaryQueue : Queue or None
                         Provide a queue here if direct access is required, 
                         otherwise one will be created and used internally
        cameraID   : int 
                     If using an acquirer that has numbered camera IDs,
                     provide this ID here for the desired camera (default = 0)
        camArgs    : tuple
                     Provide any additional arguments required by the 
                     acquirer
    """                 
    
    def __init__(self, camName, bufferSize = 10, acquisitionLock = None, 
                 imageQueue = None, auxillaryQueue = None, cameraID = 0, **camArgs):
        
        # Caller passes the name of the camera class (which should be in a
        # module of the same name) in the variable camName. This is dynamically
        # imported here and an instance created as self.cam.
        
        # We first try looking in the cas_gui.cameras folders, otherwise
        # we rely on the caller ensuring it is in the path
        
        try:
            moduleName = "cas_gui.cameras." + camName
            camModule = importlib.import_module(moduleName)
            self.cam = getattr(camModule, camName)(**camArgs)

        except:    
            moduleName = camName
            camModule = importlib.import_module(moduleName)
            self.cam = getattr(camModule, camName)(**camArgs)
            
        self.acquisitionLock = acquisitionLock
        
        super().__init__()
                
        self.cam.open_camera(cameraID)
        self._stop_event = threading.Event()
        
        self.bufferSize = bufferSize
                
        # Main queue for images for display and processing
        self.imageQueue = imageQueue
        if self.imageQueue is None:
            self.imageQueue = multiprocessing.Queue(maxsize=self.bufferSize)
        self.numDroppedFrames = 0
            
        # Second queue that is used for recording stacks etc.    
        self.auxillaryQueue = auxillaryQueue    
        if self.auxillaryQueue is None:
            self.auxillaryQueue = queue.Queue(maxsize=self.bufferSize)    
            
        self.numAuxillaryDroppedFrames = 0
        self.useAuxillaryQueue = False
        
        self.lastFrameTime = 0
        self.frameStepTime = 0
        self.currentFrame = None
        self.currentFrameNumber = 0
        self.isPaused = False
        self.isOpen = True
        self.numRemoveWhenFull = 1
        
        
    def run(self):
        """ The main loop. As long as the camera is open and we are not paused, we 
        first check that the queue is not full, and if so remove some images. Then
        we grab an image from the camera and add it to the queue. We also update some 
        records for calculating frame rate.
        """
        while self.isOpen:  

            if not self.isPaused:
                
                # Handles full queue
                if self.imageQueue.full():
                    if self.acquisitionLock is not None: self.acquisitionLock.acquire()

                    # Check for full queue
                    for idx in range(self.numRemoveWhenFull):
                        if self.imageQueue.qsize() > 0:
                            self.numDroppedFrames += 1
                            temp = self.imageQueue.get()
                            
                    if self.acquisitionLock is not None: self.acquisitionLock.release()
        
                # Try to get an image. If there is no image ready, this should
                # return None.
                frame = self.cam.get_image()

                if frame is not None:

                    self.currentFrameNumber = self.currentFrameNumber + 1
        
                    # If we have a new frame, place is in main queue. If we are
                    # using an auxillary queue at the moment, we also put a copy
                    # in there
                    self.imageQueue.put(frame)
                    if self.useAuxillaryQueue and (not self.auxillaryQueue.full()):
                        self.auxillaryQueue.put(frame) 
                    
                    self.currentFrame = frame
                    self.currentFrameTime = time.perf_counter()
                    self.frameStepTime = self.currentFrameTime - self.lastFrameTime
                    self.lastFrameTime = self.currentFrameTime
                else:
                    time.sleep(0.002)
            else:
                time.sleep(0.002)
        
    def get_camera(self):
        """ Returns a reference to the camera object.
        """
        return self.cam
    
    def get_image_queue(self):
        """ Returns a reference to the image output queue.
        """
        return self.imageQueue
    
    def get_num_images_in_queue(self):
        """ Returns the number of images in the output queue.
        """
        return self.imageQueue.qsize()
    
    
    def get_num_images_in_auxillary_queue(self):
        """ Returns the number of images in the auxillary queue.
        """
        return self.auxillaryQueue.qsize()
    
    
    def is_image_ready(self):
        """ Returns True if there is at least one image in the output queue.
        """
        if self.get_num_images_in_queue() > 0:
            return True
        else:
            return False
    
    def set_auxillary_queue_size(self, size):        
        """ Recreates auxillary queue with specified buffer size
        """
        self.auxillaryQueue = queue.Queue(maxsize = size)
        
        
    def is_auxillary_image_ready(self):
        """ Returns True if there is at least one image in the auxillary queue.
        """
        if self.get_num_images_in_auxillary_queue() > 0:
            return True
        else:
            return False    
    
    def get_next_image(self):
        """ Removes the next image from the queue and returns it.
        """
        if self.is_image_ready():
            try:
                im = self.imageQueue.get()
            except queue.Empty:
                print("Error - no image")
                return None
        
            return im
        else:
            return None
        
        
    def get_next_auxillary_image(self):
        """ Removes the next image from the auxillary queue and returns it.
        """
        if self.is_auxillary_image_ready():
            try:
                im = self.auxillaryQueue.get()
            except queue.Empty:
                print("Error - no image")
                return None
        
            return im
        else:
            return None    
    
    def get_next_image_wait(self):
        """ Wait until there is an image available in the queue and then return 
        it.
        """
        
        while self.is_image_ready() is False:
            pass
        
        try:
            if self.acquisitionLock is not None: self.acquisitionLock.acquire()
            im = self.imageQueue.get()
            if self.acquisitionLock is not None: self.acquisitionLock.release()

        except queue.Empty:
            return None
        
        return im
        
    
    def get_inter_frame_time(self):
        """ Returns the time spacing between images being acquired (this should
        be inverse of frame rate).
        """
        return self.frameStepTime
    
    
    def get_actual_fps(self):
        """ Returns the observed frame rate.
        """
        if self.frameStepTime > 0:
            return (1 / self.frameStepTime)
        else:
            return 0
        
        
    def get_latest_image(self):
        """ Returns a reference to the latest image.
        """
        return self.currentFrame
    
    
    def set_num_removal_when_full(self, num):
        """ Sets the number of images to be removed from the queue when the
        queue is full. By default this is 1, i.e. 1 image will be removed. In
        some applications which acquire a sequence of images, it is usually
        desirable to remove a full sequence of images.
        """
        self.numRemoveWhenFull = num
        
    
    def set_use_auxillary_queue(self, use):
        """ Sets whether or not to populate auxillary queue.
        """
        self.useAuxillaryQueue = use
    
    
    def flush_buffer(self):
        """ Removes all images from output queue.
        """
        while not self.imageQueue.empty():
            try:
                self.imageQueue.get_nowait()  
            except:
                pass
    
    
    def flush_auxillary_buffer(self):
        """ Removes all images from auxillary queue.
        """
        while not self.auxillaryQueue.empty():
            try:
                self.auxillaryQueue.get_nowait()  
            except:
                pass   
    
                
    def pause(self):
        """ Pauses image acquisition.
        """
        self.isPaused = True
        self.flush_buffer()
        return
    

    def resume(self):
        """ Resumes a paused image acquisition.
        """
        self.isPaused = False
        return    
        
              
    def stop(self):
        """ Shuts down the thread after closing camera.
        """
        self.isOpen = False
        time.sleep(0.5)
        if self.cam is not None:
            self.cam.close_camera()
            time.sleep(0.5)

            self.cam.dispose()
            del(self.cam)
            

            


from PyQt5.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QLabel, QDoubleSpinBox

class RangeSpinBox(QWidget):
    def __init__(self, label_text="", range_min=0.0, range_max=100.0, objectName="doubleSpinBoxWidget", connect = None, step = 0.01, parent=None):
        super().__init__(parent)

        # Set object name
        self.setObjectName(objectName)

        # Create label
        self.label = QLabel(label_text)

        # Create double spin boxes
        self.lower = QDoubleSpinBox()
        self.upper = QDoubleSpinBox()

        # Set ranges and step sizes for both spinboxes
        self.lower.setRange(range_min, range_max)
        self.upper.setRange(range_min, range_max)
        self.upper.setSingleStep(step)
        self.lower.setSingleStep(step)

        # Set object names for spinboxes
        self.lower.setObjectName(f"{objectName}_lower")
        self.upper.setObjectName(f"{objectName}_upper")

        # Layouts
        self.main_layout = QVBoxLayout(self)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.spinbox_layout = QHBoxLayout()
        self.spinbox_layout.setContentsMargins(0, 0, 0, 0)

        # Add widgets to layouts
        self.main_layout.addWidget(self.label)
        self.main_layout.addLayout(self.spinbox_layout)
        self.spinbox_layout.addWidget(self.lower)
        self.spinbox_layout.addWidget(QLabel("to"))
        self.spinbox_layout.addWidget(self.upper)

        # Set the layout
        self.setLayout(self.main_layout)

        # Set connection
        if connect is not None:
            self.lower.valueChanged.connect(connect)
            self.upper.valueChanged.connect(connect)

    def set_step(self, step):
        self.upper.setSingleStep(step)
        self.lower.setSingleStep(step)

    def disable(self):
        self.lower.setEnabled(False)
        self.upper.setEnabled(False)

    def enable(self):
        self.lower.setEnabled(True)
        self.upper.setEnabled(True)


# Example usage
if __name__ == "__main__":
    import sys
    from PyQt5.QtWidgets import QApplication

    app = QApplication(sys.argv)
    
    widget = RangeSpinBox(
        label_text="Set the range:",
        range_min=0.0,
        range_max=50.0,
        objectName="customRangeWidget"
    )
    widget.show()

    sys.exit(app.exec_())



# -*- coding: utf-8 -*-
"""
CAS: Camera Acquisition System

Camera interface for Flea Cameras (may also work for other FLIR cameras).

"""

import time
import warnings

import PySpin

from cas_gui.cameras.GenericCamera import GenericCameraInterface  
 

class FleaCameraInterface(GenericCameraInterface):
    
    _rw_modes = {
        PySpin.RO: "read only",
        PySpin.RW: "read/write",
        PySpin.WO: "write only",
        PySpin.NA: "not available"
    }

    _attr_types = {
        PySpin.intfIFloat: PySpin.CFloatPtr,
        PySpin.intfIBoolean: PySpin.CBooleanPtr,
        PySpin.intfIInteger: PySpin.CIntegerPtr,
        PySpin.intfIEnumeration: PySpin.CEnumerationPtr,
        PySpin.intfIString: PySpin.CStringPtr,
    }

    _attr_type_names = {
        PySpin.intfIFloat: 'float',
        PySpin.intfIBoolean: 'bool',
        PySpin.intfIInteger: 'int',
        PySpin.intfIEnumeration: 'enum',
        PySpin.intfIString: 'string',
        PySpin.intfICommand: 'command',
    }
    
    
    def __init__(self):        
        self.system = PySpin.System.GetInstance()
        self.camList = self.system.GetCameras()
        self.nCameras = self.camList.GetSize()        
        self.cams = self.camList
            
        
    def get_camera_list(self):
        return self.system.GetCameras()
        
        
    def open_camera(self, camID):
        
        if len(self.cams) > camID:
            
            self.cam = self.cams[camID]
            
            # Initialize camera
            self.cam.Init()
    
            self.set_value('AcquisitionMode', 'Continuous')
            
            self.cam.BeginAcquisition()            
            
            self.camera_open = True
        
            return self.cam

        else:

            return False

                
    def close_camera(self):
        
        self.cam.EndAcquisition()
        self.cam.DeInit()
        
        
    def set_value(self, nodeName, value):
        """ Writes a value to node.
        """
        
        nodemap = self.cam.GetNodeMap()        
        nodemap_tldevice = self.cam.GetTLDeviceNodeMap()  
        
        # Reference to node
        node = PySpin.CEnumerationPtr(nodemap.GetNode(nodeName))

        # Check node is writeable
        if not PySpin.IsAvailable(node) or not PySpin.IsWritable(node):
            return False
        
        # Get value we want to set
        nodeEntry = node.GetEntryByName(value)
        
        # Check this value can be written
        if not PySpin.IsAvailable(nodeEntry) or not PySpin.IsReadable(nodeEntry):
            return False
        
        nodeEntryValue = nodeEntry.GetValue()

        # Set integer value from entry node as new value of enumeration node
        node.SetIntValue(nodeEntryValue)
        
        return True
    
    
    def get_nodes(self):
        
        for node in self.cam.GetNodeMap().GetNodes():
            pit = node.GetPrincipalInterfaceType()
            name = node.GetName()
            self.camera_node_types[name] = self._attr_type_names.get(pit, pit)
            if pit == PySpin.intfICommand:
                self.camera_methods[name] = PySpin.CCommandPtr(node)
            if pit in self._attr_types:
                self.camera_attributes[name] = self._attr_types[pit](node)
        
        
        
    def get_values(self, nodeName):
        """ Returns list of possible values to set for a node """
        
        nodemap = self.cam.GetNodeMap()        
        nodemap_tldevice = self.cam.GetTLDeviceNodeMap()  
        
        # Reference to node
        node = PySpin.CEnumerationPtr(nodemap.GetNode(nodeName))
        
        entryNames = []
        if hasattr(node, 'GetEntries'):
            for entry in node.GetEntries():
                entryNames.append(entry.GetName().split('_')[-1])   
            return entryNames
        else:
            return None
        
    
    def get_value(self, nodeName):
        """ Gets the value set for a node"""
        
        nodemap = self.cam.GetNodeMap()        
        nodemap_tldevice = self.cam.GetTLDeviceNodeMap()  
        
        # Reference to node
        node = PySpin.CEnumerationPtr(nodemap.GetNode(nodeName))
        
        
            
        if hasattr(node, "GetValue"):
            return node.GetValue()
        elif hasattr(node, "ToString"):
           return node.ToString()
        else:
            return None
        
        
    
    def get_nodes(self):
        """ Returns a list of nodes """
        
        nodes = []
        for node in self.cam.GetNodeMap().GetNodes():
            nodes.append(node.GetName())
         
        nodes.sort()    
        
        return nodes
        
       
    def dispose(self):
        del self.cam 
        
        # Clear camera list before releasing system
        self.camList.Clear()

        # Release system instance
        self.system.ReleaseInstance()
        
        
               
    def get_image(self, timeout = 10000):
        try:
            image = self.cam.GetNextImage(timeout)
        except:
            image = None
        
        if  image is None:
            return None
        else:
            imageData = image.GetNDArray()
            return imageData
    
    
    
    
    ###### Frame Rate

    def set_frame_rate_on(self):
        """ Enables or disables frame rate control """

        
        try:
            nodemap = self.cam.GetNodeMap()
            node = PySpin.CEnumerationPtr(nodemap.GetNode("AcquisitionFrameRateAuto"))
            node.SetIntValue(0)
            return True  
        except:
            return False
        
        
    def set_trigger_mode(self, enabled):
        """ Enables or disables trigger mode """
        
        if enabled:
            return self.set_value('TriggerMode', 'On')
        else:
            return self.set_value('TriggerMode', 'Off')


    def set_frame_rate(self, fps):
                
        # Make sure auto frame rate is disabled, will return false if can't turn off 
        # auto frame rate
        success = self.set_frame_rate_on()
        
        try:                
            if self.cam.AcquisitionFrameRate.GetAccessMode() == PySpin.RW:
                self.cam.AcquisitionFrameRate.SetValue(fps)
                return True
        except:
                print("Can't set frame rate, in automatic frame rate mode.")
                return False    
    
    
    def get_frame_rate(self):
        
        if self.is_frame_rate_enabled():
            return self.cam.AcquisitionFrameRate.GetValue()
        else:
            return 0

    
    def get_frame_rate_range(self):
        
        return self.cam.AcquisitionFrameRate.GetMin(), self.cam.AcquisitionFrameRate.GetMax()      
    
    
    def is_frame_rate_enabled(self):
        
        return (self.cam.AcquisitionFrameRate.GetAccessMode() == PySpin.RW )
    
       
    def get_measured_frame_rate(self):
        
        return None 


    ##### Exposure
    def is_exposure_enabled(self):
        return False

    def set_exposure(self, exposure):
        
        # Ensure auto exposure is off
        self.cam.ExposureAuto.SetValue(0)
        
        if exposure < self.get_exposure_range()[0]: 
            exposure = self.get_exposure_range()[0]
            warnings.warn(f"Set exposure limited to minimum allowed value of {self.get_exposure_range()[0]}.")
        if exposure > self.get_exposure_range()[1]: 
            exposure = self.get_exposure_range()[1]
            warnings.warn(f"Set exposure limited to maximum allowed value of {self.get_exposure_range()[1]}.")
        
        if self.cam.ExposureTime.GetAccessMode() == PySpin.RW:
            try:
                self.cam.ExposureTime.SetValue(exposure)
                return True
            except:
                return False
                print(f"FleaCameraInterface: Cannot set exposure value of {exposure} us.")
        else:
            return False             
        
    def get_exposure(self):
        return self.cam.ExposureTime.GetValue() 
    
    def get_exposure_range(self):
        return self.cam.ExposureTime.GetMin(), self.cam.ExposureTime.GetMax()
 
    
        
    ##### Gain    
    def is_gain_enabled(self):
        return False
        
    
    def set_gain(self, gain):
        
        # Ensure auto gain is off
        self.cam.GainAuto.SetValue(0)  
        
        if gain < self.get_gain_range()[0]: 
            gain = self.get_gain_range()[0]
            warnings.warn(f"Set gain limited to minimum allowed value of {self.get_gain_range()[0]}.")
        if gain > self.get_gain_range()[1]: 
            gain = self.get_gain_range()[1]
            warnings.warn(f"Set gain limited to maximum allowed value of {self.get_gain_range()[1]}.")

        if self.cam.Gain.GetAccessMode() == PySpin.RW:
            try:
                self.cam.Gain.SetValue(gain)
                return True
            except:
                return False

            return True
        else:
            return False 
        

    
    def get_gain(self):
        return self.cam.Gain.GetValue() 
    
    
    def get_gain_range(self):
        return self.cam.Gain.GetMin(), self.cam.Gain.GetMax()
        
        

if __name__ == "__main__":
    
    ### Performs a test acquisition
    cams = FleaCameraInterface()
    cams.init()
    cam = cams.open(0)
    
    if cam != False:
    
        cam.set_frame_rate(60)
        print(f"Exposure: {cam.get_exposure()}")
        print(f"Frame Rate: {cam.get_frame_rate()}")
        t1 = time.time()
        for i in range(120):
            im = cam.get_image()
        t2 = time.time()
        print(f"Time to get 120 frames: {t2-t1}")
        cam.dispose()  
        del(cam)
    
    cams.deInit()



# -*- coding: utf-8 -*-
"""
CAS: Camera Acquisition System

A camera interface for CAS which instead of using a camera, loads images 
from a file (single or stack).

"""
import time

from PIL import Image, ImageSequence
import numpy as np

from cas_gui.cameras.GenericCamera import GenericCameraInterface


class FileInterface(GenericCameraInterface):
        
    dtype = 'uint16'
    currentImageIdx = 0
    lastImageIdx = -1
    lastImage = None
    fileOpen = False
    dataset = None
    stack = None
    
    def __init__(self, **kwargs): 
        
        self.filename = kwargs.get('filename', None)
        self.lastImageTime = time.perf_counter()
        self.fps = 30
        self.frameRateEnabled = False

        if self.filename is not None:
            try:
                self.dataset = Image.open(self.filename)
                self.fileOpen = True
                self.stack = None
            except:
                self.fileOpen = False
                return None
        else:
            self.fileOpen = False
            return None
                   
        
    def __str__(self):
        return "File Processor, source = " + self.filename  
    

    def is_file_open(self):
        return self.fileOpen
    
    def get_camera_list(self):
        return "File Processor, source = " + self.filename  
    
        
    def open_camera(self, camID): 
        pass       
    
                
    def close_camera(self):
        pass
    
        
    def dispose(self):
        self.dataset.close()       
      
               
    def get_image(self):
        """ Get the desired image from the file.
        """

        # If we have already loaded this image, just return it
        if self.currentImageIdx == self.lastImageIdx:
            return self.lastImage
            
        # Otherwise jump to desired image (if it is a stack)
        self.dataset.seek(self.currentImageIdx)
        imData = np.asarray(self.dataset)

       
        # Store this image in lastImage
        self.lastImage = imData
        self.lastImageIdx = self.currentImageIdx

       
        return imData
    
    
    def get_all_images(self):
        
        if self.stack is None:
        
            with Image.open(self.filename) as im:
            
                if im.n_frames > 1:
                    h,w = np.shape(im)
                    dt = np.array(im).dtype
                    self.stack = np.zeros((im.n_frames, h,w), dtype = dt)
                    for i in range(im.n_frames):
                        im.seek(i)
                        self.stack[i,:,:] = np.array(im)
                    return self.stack
                else:    
                    return np.array(Image.open(self.filename))
        else:
            return self.stack
    
    def set_image_idx(self, idx):
        """ If the file contains multiple frames, sets the index of the
        image that will be returned when get_image is called.
        """
        self.currentImageIdx = idx
        
        
    def get_number_images(self):
        """ If the file contains multiple frames, returns the number of frames.
        Will return 1 if a singe image.
        """
        if self.dataset is not None:
            return self.dataset.n_frames
        
    
    ###### Frame Rate
    def enable_frame_rate(self):
        self.frameRateEnabled = False
        return False    

    def disable_frame_rate(self):
        self.frameRateEnabled = False
        return True     

    def set_frame_rate(self, fps):
        pass    

    def get_frame_rate(self):
        return None
    
    def get_frame_rate_range(self):
        return None        
    
    def is_frame_rate_enabled(self):
        return False
    
      


    ##### Exposure
    def is_exposure_enabled(self):
        return False

    def set_exposure(self, exposure):
        pass      
        
    def get_exposure(self):
        return 0
    
    def get_exposure_range(self):
        return 0,0

    
        
    ##### Gain    
    def isGainEnabled(self):
        return False
        
    def setGain(self, gain):
        pass
    
    def getGain(self):
        return 0
    
    def getGainRange(self):
        return 0,0
        
        

if __name__ == "__main__":
    print("Test mode not implemented")